/*--- formatted by Jindent 2.1, (www.c-lab.de/~jindent) ---*/

/*
 * JPython Database Specification API 2.0
 * 
 * $Id: PyConnection.java,v 1.4 2000/07/06 00:24:02 bzimmer Exp $
 * 
 * Copyright (C) 2000 brian zimmer <bzimmer@ziclix.com>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 */
package com.ziclix.python.sql;

import java.sql.*;
import org.python.core.*;

/**
 * A connection to the database.
 * 
 * @author brian zimmer
 * @author last revised by $Author: bzimmer $
 * @version $Revision: 1.4 $
 */
public class PyConnection extends PyObject implements InitModule {
	private Connection connection;
	private boolean supportsTransactions;

	// __class__ boilerplate -- see PyObject for details
	public static PyClass __class__;

	/**
	 * Don't really want this to be public but InitModule requires it.
	 * 
	 */
	public PyConnection() {
		this.connection = null;
		this.supportsTransactions = false;
	}

	/**
	 * Set the JDBC connection.
	 * 
	 * @param connection
	 */
	void setConnection(Connection connection) throws SQLException {
		this.connection = connection;
		this.supportsTransactions = this.connection.getMetaData().supportsTransactions();

		if(this.supportsTransactions) {
			this.connection.setAutoCommit(false);
		} 
	} 

	/**
	 * Produces a string representation of the object.
	 * 
	 * @return string representation of the object.
	 */
	public String toString() {
		try {
			return "<PyConnection user='" + this.connection.getMetaData().getUserName() + "', url='" 
				   + this.connection.getMetaData().getURL() + "'>";
		} catch(SQLException e) {
			return "<PyConnection at " + hashCode() + ">";
		} 
	} 

	/**
	 * $METHOD_DESCRIPTION$
	 * 
	 * @param dict
	 */
	public void initModule(PyObject dict) {
		dict.__setitem__("autocommit", new PyInteger(0));
		dict.__setitem__("__version__", 
				Py.newString("$Revision: 1.4 $").__getslice__(Py.newInteger(11), Py.newInteger(-2), null));
		dict.__setitem__("close", new ConnectionFunc("close", 0, 0, 0, "close the connection"));
		dict.__setitem__("commit", new ConnectionFunc("commit", 1, 0, 0, "commit the connection"));
		dict.__setitem__("cursor", new ConnectionFunc("cursor", 2, 0, 0, "obtain a cursor"));
		dict.__setitem__("rollback", new ConnectionFunc("rollback", 3, 0, 0, "rollback the connection"));

		// hide from python
		dict.__setitem__("initModule", null);
		dict.__setitem__("__class__", null);
	} 

	/**
	 * $METHOD_DESCRIPTION$
	 * 
	 * @param name
	 * @param value
	 */
	public void __setattr__(String name, PyObject value) {
		if("autocommit".equals(name)) {
			try {
				boolean autocommit = ((PyInteger) value).getValue() == 0 ? false : true;
				if(this.supportsTransactions) {
					this.connection.setAutoCommit(autocommit);
				} 
			} catch(SQLException e) {
				throw zxJDBC.DatabaseError(e.getMessage());
			} 

			return;
		} 

		super.__setattr__(name, value);
	} 

	/**
	 * $METHOD_DESCRIPTION$
	 * 
	 * @param name
	 * @return
	 */
	public PyObject __findattr__(String name) {
		if("autocommit".equals(name)) {
			try {
				return connection.getAutoCommit() ? Py.One : Py.Zero;
			} catch(SQLException e) {
				throw zxJDBC.DatabaseError(e.getMessage());
			} 
		} else if("dbname".equals(name)) {
			try {
				return Py.newString(this.connection.getMetaData().getDatabaseProductName());
			} catch(SQLException e) {
				throw zxJDBC.DatabaseError(e.getMessage());
			} 
		} else if("dbversion".equals(name)) {
			try {
				return Py.newString(this.connection.getMetaData().getDatabaseProductVersion());
			} catch(SQLException e) {
				throw zxJDBC.DatabaseError(e.getMessage());
			} 
		} 

		return super.__findattr__(name);
	} 

	/**
	 * $METHOD_DESCRIPTION$
	 * 
	 * @return
	 */
	protected PyClass getPyClass() {
		return __class__;
	} 

	/**
	 * Close the connection now (rather than whenever __del__ is called).
	 * The connection will be unusable from this point forward; an Error
	 * (or subclass) exception will be raised if any operation is attempted
	 * with the connection. The same applies to all cursor objects trying
	 * to use the connection.
	 * 
	 */
	public void close() {
		try {

			// close the cursors too?
			connection.close();
		} catch(SQLException e) {
			throw zxJDBC.DatabaseError(e.getMessage());
		} 
	} 

	/**
	 * Commit any pending transaction to the database. Note that if the
	 * database supports an auto-commit feature, this must be initially
	 * off. An interface method may be provided to turn it back on.
	 * 
	 * Database modules that do not support transactions should implement
	 * this method with void functionality.
	 * 
	 */
	public void commit() {
		if(!this.supportsTransactions) {
			return;
		} 

		try {
			connection.commit();
		} catch(SQLException e) {
			throw zxJDBC.DatabaseError(e.getMessage());
		} 
	} 

	/**
	 * <i>This method is optional since not all databases provide transaction
	 * support.</i>
	 * 
	 * In case a database does provide transactions this method causes the database
	 * to roll back to the start of any pending transaction. Closing a connection
	 * without committing the changes first will cause an implicit rollback to be
	 * performed.
	 * 
	 */
	public void rollback() {
		if(!this.supportsTransactions) {
			return;
		} 

		try {
			connection.rollback();
		} catch(SQLException e) {
			throw zxJDBC.DatabaseError(e.getMessage());
		} 
	} 

	/**
	 * Return a new Cursor Object using the connection. If the database does not
	 * provide a direct cursor concept, the module will have to emulate cursors
	 * using other means to the extent needed by this specification.
	 * 
	 * @return a new cursor using this connection
	 */
	public PyCursor cursor() {
		PyCursor c = new PyExtendedCursor();
		c.setConnection(this.connection);

		return c;
	} 

	/**
	 * $CLASS_DESCRIPTION$
	 * 
	 * @author
	 * @author last revised by $Author: bzimmer $
	 * @version $Revision: 1.4 $
	 */
	class ConnectionFunc extends PyBuiltinFunctionSet {

		/**
		 * $CONSTRUCTOR_DESCRIPTION$
		 * 
		 * @param name
		 * @param index
		 * @param minargs
		 * @param maxargs
		 * @param doc
		 */
		ConnectionFunc(String name, int index, int minargs, int maxargs, String doc) {
			super(name, index, minargs, maxargs, true, doc);
		}

		/**
		 * $METHOD_DESCRIPTION$
		 * 
		 * @return
		 */
		public PyObject __call__() {
			PyConnection c = (PyConnection) __self__;
			switch(index) {

			case 0:
				c.close();

				return Py.None;

			case 1:
				c.commit();

				return Py.None;

			case 2:
				return c.cursor();

			case 3:
				c.rollback();

				return Py.None;

			default:
				throw argCountError(0);
			}
		} 
	}

}

