/*
 * JPython Database Specification API 2.0
 *
 * $Id: PyCursor.java,v 1.11 2000/07/06 19:02:42 bzimmer Exp $
 *
 * Copyright (C) 2000 brian zimmer <bzimmer@ziclix.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */
package com.ziclix.python.sql;

import java.io.*;
import java.sql.*;
import java.math.*;
import java.util.*;
import org.python.core.*;

/**
 * These objects represent a database cursor, which is used to manage the
 * context of a fetch operation.
 *
 * @author brian zimmer
 * @author last revised by $Author: bzimmer $
 * @version $Revision: 1.11 $
 */
public class PyCursor extends PyObject implements InitModule {
	protected Connection connection;
	protected PyObject results;
	protected int counter;
	protected int rowcount;
	protected PyObject description;
	protected int arraysize;
	protected Hashtable cache;
	protected DataHandler datahandler;
	private static boolean jdbc2 = false;
	static {
		try {
			Class.forName("java.sql.Clob");

			jdbc2 = true;
		} catch(Throwable e) {}
	}

	/**
	 * InitModule requires there be an empty constructor, but I really
	 * want the Statement to be passed in the constructor.  What to do?
	 */
	public PyCursor() {
		this.description = Py.None;
		this.rowcount = -1;
		this.cache = new Hashtable(10);
		this.arraysize = 1;
		this.connection = null;

		if(jdbc2) {
			this.datahandler = new JDBC20DataHandler();
		} else {
			this.datahandler = new DefaultDataHandler();
		}
	}

	// __class__ boilerplate -- see PyObject for details
	public static PyClass __class__;

	/**
	 * $METHOD_DESCRIPTION$
	 *
	 * @return
	 */
	protected PyClass getPyClass() {
		return __class__;
	}

	protected static PyList __methods__;
	static {
		PyObject[] m = {
			new PyString("close"), new PyString("execute"), new PyString("executemany"), new PyString("fetchone"),
			new PyString("fetchall"), new PyString("fetchmany"), new PyString("callproc")
		};
		__methods__ = new PyList(m);
	}

	protected static PyList __members__;
	static {
		PyObject[] m = {
			new PyString("arraysize"), new PyString("rowcount"), new PyString("description"),
			new PyString("datahandler")
		};
		__methods__ = new PyList(m);
	}

	/**
	 * String representation of the object.
	 *
	 * @return a string representation of the object.
	 */
	public String toString() {
		return "<PyCursor object instance at " + hashCode() + ">";
	}

	/**
	 * Sets the connection.
	 *
	 * @param connection a JDBC connection
	 */
	void setConnection(Connection connection) {
		this.connection = connection;
	}

	/**
	 * Sets the attribute name to value.
	 *
	 * @param name
	 * @param value
	 */
	public void __setattr__(String name, PyObject value) {
		if("arraysize".equals(name)) {
			arraysize = ((PyInteger) value).getValue();
		} else if("datahandler".equals(name)) {
			this.datahandler = (DataHandler) value.__tojava__(DataHandler.class);
		} else {
			super.__setattr__(name, value);
		}
	}

	/**
	 * Gets the value of the attribute name.
	 *
	 * @param name
	 * @return
	 */
	public PyObject __findattr__(String name) {
		if("arraysize".equals(name)) {
			return new PyInteger(arraysize);
		} else if("rowcount".equals(name)) {
			return new PyInteger(rowcount);
		} else if("description".equals(name)) {
			return description;
		} else if("__methods__".equals(name)) {
			return __methods__;
		} else if("__members__".equals(name)) {
			return __members__;
		} else if("datahandler".equals(name)) {
			return Py.java2py(this.datahandler);
		}

		return super.__findattr__(name);
	}

	/**
	 * Initializes the object's namespace.
	 *
	 * @param dict
	 */
	public void initModule(PyObject dict) {
		dict.__setitem__("__version__",
				Py.newString("$Revision: 1.11 $").__getslice__(Py.newInteger(11), Py.newInteger(-2), null));
		dict.__setitem__("arraysize", new PyInteger(arraysize));
		dict.__setitem__("rowcount", new PyInteger(rowcount));
		dict.__setitem__("description", this.description);
		dict.__setitem__("datahandler", Py.java2py(this.datahandler));
		dict.__setitem__("fetchmany", new CursorFunc("fetchmany", 0, 1, 2, "fetch specified number of rows"));
		dict.__setitem__("close", new CursorFunc("close", 1, 1, "close the cursor"));
		dict.__setitem__("fetchall", new CursorFunc("fetchall", 2, 1, "fetch all results"));
		dict.__setitem__("fetchone", new CursorFunc("fetchone", 3, 1, "fetch the next result"));
		dict.__setitem__("nextset", new CursorFunc("nextset", 4, 1, "return next set or None"));
		dict.__setitem__("execute", new CursorFunc("execute", 5, 1, 3, "execute the sql expression"));
		dict.__setitem__("setinputsizes", new CursorFunc("setinputsizes", 6, 1, "not implemented"));
		dict.__setitem__("setoutputsize", new CursorFunc("setoutputsize", 7, 1, 2, "not implemented"));
		dict.__setitem__("callproc", new CursorFunc("callproc", 8, 1, 2, "executes a stored procedure"));
		dict.__setitem__("executemany", new CursorFunc("executemany", 9, 1, 3, "execute sql with the parameter list"));

		// hide from python
		dict.__setitem__("initModule", null);
		dict.__setitem__("__del__", null);
		dict.__setitem__("__setattr__", null);
		dict.__setitem__("__class__", null);
	}

	/**
	 * Delete the cursor.
	 *
	 */
	public void __del__() {
		close();
	}

	/**
	 * Close the cursor now (rather than whenever __del__ is called).
	 * The cursor will be unusable from this point forward; an Error
	 * (or subclass) exception will be raised if any operation is
	 * attempted with the cursor.
	 *
	 */
	public void close() {}

	/**
	 * Create a Statement.
	 *
	 * @return
	 */
	protected Statement statement() throws SQLException {
		return this.connection.createStatement();
	}

	/**
	 * Call a stored procedure.
	 *
	 * @param sqlString
	 * @return
	 * @throws SQLException
	 */
	protected CallableStatement callableStatement(String sqlString) throws SQLException {
		return this.connection.prepareCall(sqlString);
	}

	/**
	 * Prepare a statement ready for executing.
	 *
	 * @param sqlString
	 * @return
	 * @throws SQLException
	 */
	protected PreparedStatement prepareStatement(String sqlString) throws SQLException {
		PreparedStatement stmt = null;
		if(cache.containsKey(sqlString)) {
			stmt = (PreparedStatement) cache.get(sqlString);

			stmt.clearParameters();
		} else {
			stmt = this.connection.prepareStatement(sqlString);

			cache.put(sqlString, stmt);
		}

		return stmt;
	}

	/**
	 * This method is optional since not all databases provide stored procedures.
	 *
	 * Call a stored database procedure with the given name. The sequence of parameters
	 * must contain one entry for each argument that the procedure expects. The result of
	 * the call is returned as modified copy of the input sequence. Input parameters are
	 * left untouched, output and input/output parameters replaced with possibly new values.
	 *
	 * The procedure may also provide a result set as output. This must then be made available
	 * through the standard fetchXXX() methods.
	 */
	public void callproc(String procName, PyObject params) {
		boolean supported = false;
		clear();

		try {
			supported = this.connection.getMetaData().supportsStoredProcedures();
		} catch(SQLException e) {}

		if(!supported) {
			throw zxJDBC.NotSupportedError("callproc unsupported");
		} else {
			throw zxJDBC.NotSupportedError("stored procedures not implemented in db");
		}
	}

	/**
	 * Prepare a database operation (query or command) and then execute it against all
	 * parameter sequences or mappings found in the sequence seq_of_parameters.
	 * Modules are free to implement this method using multiple calls to the execute()
	 * method or by using array operations to have the database process the sequence as
	 * a whole in one call.
	 *
	 * The same comments as for execute() also apply accordingly to this method.
	 *
	 * Return values are not defined.
	 */
	public void executemany(String sqlString, PyObject params, PyObject bindings) {
		clear();

		if(PyUtil.PyTupleList_Check(params)) {
			for(int i = 0; i < params.__len__(); i++) {
				PyObject obj = params.__getitem__(i);
				if(!PyUtil.PyTupleList_Check(obj)) {
					throw zxJDBC.InterfaceError("argument must be a list or tuple");
				} else {
					execute(sqlString, obj, bindings);
				}
			}
		} else {
			throw zxJDBC.InterfaceError("second argument must be a list or tuple of lists or tuples");
		}
	}

	/**
	 * Prepare and execute a database operation (query or command).
	 * Parameters may be provided as sequence or mapping and will
	 * be bound to variables in the operation. Variables are specified
	 * in a database-specific notation (see the module's paramstyle
	 * attribute for details).
	 *
	 * A reference to the operation will be retained by the cursor.
	 * If the same operation object is passed in again, then the cursor
	 * can optimize its behavior. This is most effective for algorithms
	 * where the same operation is used, but different parameters are
	 * bound to it (many times).
	 *
	 * For maximum efficiency when reusing an operation, it is best to
	 * use the setinputsizes() method to specify the parameter types and
	 * sizes ahead of time. It is legal for a parameter to not match the
	 * predefined information; the implementation should compensate, possibly
	 * with a loss of efficiency.
	 *
	 * The parameters may also be specified as list of tuples to e.g. insert
	 * multiple rows in a single operation, but this kind of usage is
	 * deprecated: executemany() should be used instead.
	 *
	 * Return values are not defined.
	 *
	 * @param sqlString sql string
	 * @param params params for a prepared statement
	 * @param bindings dictionary of (param index : SQLType binding)
	 */
	public void execute(String sqlString, PyObject params, PyObject bindings) {
		clear();

		try {
			if(params == null || params.equals(Py.None) || params.__len__() == 0) {
				Statement stmt = statement();
				boolean flag = stmt.execute(sqlString);
				create(stmt.getResultSet());
			} else {
				if(!PyUtil.PyTupleList_Check(params)) {
					throw zxJDBC.InterfaceError("optional second argument must be a list or tuple");
				}

				if(PyUtil.PyTupleListOfTupleList_Check(params)) {
					for(int i = 0; i < params.__len__(); i++) {
						execute(sqlString, params.__getitem__(i), bindings);
					}
				} else {
					PreparedStatement stmt = prepareStatement(sqlString);
					int param_count = params.__len__();
					boolean has_bindings = (bindings != null &&!bindings.equals(Py.None)
							&& bindings instanceof PyDictionary);
					for(int i = 0; i < param_count; i++) {
						if(has_bindings) {
							PyDictionary d = (PyDictionary) bindings;
							PyObject index = new PyInteger(i);
							if(d.has_key(index)) {
								PyObject k = d.__getitem__(index);
								if(k != null &&!Py.None.equals(k)) {
									int type = ((Integer) k.__tojava__(Integer.class)).intValue();
									this.datahandler.setJDBCObject(stmt, i + 1, params.__getitem__(i), type);
								}
							} else {
								this.datahandler.setJDBCObject(stmt, i + 1, params.__getitem__(i));
							}
						} else {
							this.datahandler.setJDBCObject(stmt, i + 1, params.__getitem__(i));
						}
					}

					boolean flag = stmt.execute();
					create(stmt.getResultSet());
				}
			}
		} catch(SQLException e) {
			throw zxJDBC.Error(e);
		}
	}

	/**
	 * Fetch the next row of a query result set, returning a single sequence,
	 * or None when no more data is available.
	 *
	 * An Error (or subclass) exception is raised if the previous call to
	 * executeXXX() did not produce any result set or no call was issued yet.
	 *
	 * @return a single sequence from the result set, or None when no more data is available
	 */
	public PyObject fetchone() {
		return fetchmany(1);
	}

	/**
	 * Fetch all (remaining) rows of a query result, returning them as a sequence
	 * of sequences (e.g. a list of tuples). Note that the cursor's arraysize attribute
	 * can affect the performance of this operation.
	 *
	 * An Error (or subclass) exception is raised if the previous call to executeXXX()
	 * did not produce any result set or no call was issued yet.
	 *
	 * @return a sequence of sequences from the result set, or None when no more data is available
	 */
	public PyObject fetchall() {
		return fetchmany(this.rowcount);
	}

	/**
	 * Fetch the next set of rows of a query result, returning a sequence of
	 * sequences (e.g. a list of tuples). An empty sequence is returned when
	 * no more rows are available.
	 *
	 * The number of rows to fetch per call is specified by the parameter. If
	 * it is not given, the cursor's arraysize determines the number of rows
	 * to be fetched. The method should try to fetch as many rows as indicated
	 * by the size parameter. If this is not possible due to the specified number
	 * of rows not being available, fewer rows may be returned.
	 *
	 * An Error (or subclass) exception is raised if the previous call to executeXXX()
	 * did not produce any result set or no call was issued yet.
	 *
	 * Note there are performance considerations involved with the size parameter.
	 * For optimal performance, it is usually best to use the arraysize attribute.
	 * If the size parameter is used, then it is best for it to retain the same value
	 * from one fetchmany() call to the next.
	 *
	 * @return a sequence of sequences from the result set, or None when no more data is available
	 */
	public PyObject fetchmany(int size) {
		PyObject res = Py.None;
		if(size <= 0) {
			size = this.rowcount;
		}

		if((counter + 1) < this.rowcount) {
			int start = counter + 1;
			counter += size;
			res = results.__getslice__(new PyInteger(start), new PyInteger(counter + 1), new PyInteger(1));
		}

		return res;
	}

	/**
	 * Returns the next set, else None if no more sets exist.
	 *
	 * @return
	 */
	public PyObject nextset() {
		return Py.None;
	}

	/**
	 * Create the results after a successful execution.
	 *
	 * @param rs A ResultSet.
	 */
	protected void create(ResultSet rs) {
		try {
			if(rs != null && rs.getMetaData() != null) {
				this.description = createDescription(rs.getMetaData());
				this.results = createResults(rs);
				this.rowcount = this.results.__len__();
			}
		} catch(SQLException e) {
			throw zxJDBC.Error(e);
		}
		finally {
			if(rs != null) {
				try {
					rs.close();
				} catch(SQLException e) {
					throw zxJDBC.Error(e);
				}
			}
		}
	}

	/**
	 * (name, type_code, display_size, internal_size, precision, scale, null_ok)
	 */
	protected PyList createDescription(ResultSetMetaData meta) throws SQLException {
		PyList des = new PyList();
		for(int i = 1; i <= meta.getColumnCount(); i++) {
			PyObject[] a = new PyObject[7];
			a[0] = new PyString(meta.getColumnName(i));
			a[1] = new PyInteger(meta.getColumnType(i));	// should figure out the correct python type here
			a[2] = new PyInteger(meta.getColumnDisplaySize(i));
			a[3] = Py.None;
			a[4] = new PyInteger(meta.getPrecision(i));
			a[5] = new PyInteger(meta.getScale(i));
			a[6] = new PyInteger(meta.isNullable(i));

			des.append(new PyTuple(a));
		}

		return des;
	}

	/**
	 * Creates the results of a query.  Iterates through the list and builds the tuple.
	 *
	 * @param set
	 * @return a list of tuples of the results
	 * @throws Exception
	 */
	protected PyList createResults(ResultSet set) throws SQLException {
		PyList res = new PyList();
		int desc_len = description.__len__();
		while(set.next()) {
			PyObject[] l = new PyObject[desc_len];
			for(int i = 0; i < desc_len; i++) {
				int type = ((PyInteger) description.__getitem__(i).__getitem__(1)).getValue();
				l[i] = this.datahandler.getPyObject(set, i + 1, type);
			}

			res.append(new PyTuple(l));
		}

		return res;
	}

	/**
	 * Reset the cursor.
	 *
	 */
	protected void clear() {
		this.description = Py.None;
		this.results = Py.None;
		this.rowcount = -1;
		this.counter = this.rowcount;
	}

	/**
	 * $CLASS_DESCRIPTION$
	 *
	 * @author
	 * @author last revised by $Author: bzimmer $
	 * @version $Revision: 1.11 $
	 */
	class CursorFunc extends PyBuiltinFunctionSet {

		/**
		 * $CONSTRUCTOR_DESCRIPTION$
		 *
		 * @param name
		 * @param index
		 * @param argcount
		 * @param doc
		 */
		CursorFunc(String name, int index, int argcount, String doc) {
			super(name, index, argcount, argcount, true, doc);
		}

		/**
		 * $CONSTRUCTOR_DESCRIPTION$
		 *
		 * @param name
		 * @param index
		 * @param minargs
		 * @param maxargs
		 * @param doc
		 */
		CursorFunc(String name, int index, int minargs, int maxargs, String doc) {
			super(name, index, minargs, maxargs, true, doc);
		}

		/**
		 * $METHOD_DESCRIPTION$
		 *
		 * @return
		 */
		public PyObject __call__() {
			PyCursor cursor = (PyCursor) __self__;
			switch(index) {

			case 0:
				return cursor.fetchmany(cursor.arraysize);

			case 1:
				cursor.close();

				return Py.None;

			case 2:
				return cursor.fetchall();

			case 3:
				return cursor.fetchone();

			case 4:
				return cursor.nextset();

			default:
				throw argCountError(0);
			}
		}

		/**
		 * $METHOD_DESCRIPTION$
		 *
		 * @param arg
		 * @return
		 */
		public PyObject __call__(PyObject arg) {
			PyCursor cursor = (PyCursor) __self__;
			switch(index) {

			case 0:
				return cursor.fetchmany(((PyInteger) arg).getValue());

			case 5:
				cursor.execute((String) arg.__tojava__(String.class), Py.None, Py.None);

				return Py.None;

			case 6:

			case 7:
				return Py.None;

			case 8:
				cursor.callproc((String) arg.__tojava__(String.class), Py.None);

				return Py.None;

			case 9:
				cursor.executemany((String) arg.__tojava__(String.class), Py.None, Py.None);

				return Py.None;

			default:
				throw argCountError(1);
			}
		}

		/**
		 * $METHOD_DESCRIPTION$
		 *
		 * @param arga
		 * @param argb
		 * @return
		 */
		public PyObject __call__(PyObject arga, PyObject argb) {
			PyCursor cursor = (PyCursor) __self__;
			switch(index) {

			case 5:
				cursor.execute((String) arga.__tojava__(String.class), argb, Py.None);

				return Py.None;

			case 7:
				return Py.None;

			case 8:
				cursor.callproc((String) arga.__tojava__(String.class), argb);

				return Py.None;

			case 9:
				cursor.executemany((String) arga.__tojava__(String.class), argb, Py.None);

				return Py.None;

			default:
				throw argCountError(2);
			}
		}

		/**
		 * $METHOD_DESCRIPTION$
		 *
		 * @param arga
		 * @param argb
		 * @return
		 */
		public PyObject __call__(PyObject arga, PyObject argb, PyObject argc) {
			PyCursor cursor = (PyCursor) __self__;
			switch(index) {

			case 5:
				cursor.execute((String) arga.__tojava__(String.class), argb, argc);

				return Py.None;

			case 9:
				cursor.executemany((String) arga.__tojava__(String.class), argb, argc);

				return Py.None;

			default:
				throw argCountError(2);
			}
		}
	}

}

